import asyncio
import logging
import sys
import time
import signal
import os

import pytest

from pyskyq.asyncthread import AsyncThread

logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
logging.basicConfig(level=logging.WARNING, stream=sys.stdout,
                    format=logformat)  # datefmt="%Y-%m-%d %H:%M:%S"

LOGGER = logging.getLogger(__name__)


def test_asyncthread():

    at1 = AsyncThread()
    at2 = AsyncThread()

    # test singleton.
    assert at1.loop is at2.loop
    assert at1.thread is at2.thread

    assert at1.shutdown_sentinel is False
    assert at1.thread.is_alive()

    async def t1():
        await asyncio.sleep(1.2)
        return "He's not the Messiah..."

    async def t2():
        await asyncio.sleep(1.4)
        return "...he's very naughty boy."


    async def cancel_me():
        try:
            # Wait for 1 hour
            await asyncio.sleep(3600)
        except asyncio.CancelledError:
            await asyncio.sleep(5)
            raise

    f1 = at1.run(t1())
    f2 = at1.run(t2())
    at1.run(cancel_me())

    while f1.running():
        time.sleep(.1)
    assert f1.result() == "He's not the Messiah..."

    while f1.running():
        time.sleep(.1)
    assert f2.result() == "...he's very naughty boy."

    at1.shutdown()
    assert at1.shutdown_sentinel is True

    time.sleep(0.1)
    assert not at1.loop.is_running()

    at3 = AsyncThread()  #create a new one to test re-initing...
    assert at3.shutdown_sentinel is False
    assert at3.thread.is_alive()




"""This module implements an asyncio-based cron-related classes.

It is heavily based on (ie, copied from):
    https://github.com/gawel/aiocron

Credit to Gael Pasgrimaud for his work here.

"""
import asyncio
import threading
import time
import aiocron


class CronThread(threading.Thread):
    """Implement a cronjob in a thread."""

    def __init__(self):  # set up the thread
        """Initialise the thread."""
        super(CronThread, self).__init__()
        self.loop = None
        self.start()
        time.sleep(.1)

    def run(self): # the thread code itself.
        """Get an event loop and start it up in this thread."""
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()
        self.loop.close()

    def stop(self):
        """Stop the loop and join the thread."""
        self.loop.call_soon_threadsafe(self.loop.stop)
        self.join()

    def crontab(self, *args, **kwargs):
        """Return an aiocron.contab object running in this loop."""
        kwargs['loop'] = self.loop
        return aiocron.crontab(*args, **kwargs)



# pylint: disable=no-member
"""This module provides global management of asyncio-based tasks in a dedicated thread.

It provides the :class:`AsyncThread` class which is a singleton with some
specical quirks. When instantiated it will figure out if it has been
instantiated before. If this is the first instantiation, it will create an
internal thread and run an event loop in it. If it's being instantiated again,
it will check the state of the thead and the event loop within it and restart
those if needed. It also provides signal handlers that cleanly cancel and await
the completion of any tasks upon receipt of a ``SIGINT`` or ``SIGTERM``
interrupt.
"""
import asyncio
import threading
import signal
import logging
import functools
from concurrent.futures import Future
from enum import Enum
from typing import Coroutine

LOGGER = logging.getLogger(__name__)

class AsyncThread():
    """This class holds references to a separate thread which runs an asyncio event loop.

    It is a Singleton.

    Attributes:
        loop (asyncio.AbstractEventLoop): The associated event loop.
        thread (threading.Thread): The thread that the event loop is running in.

    Example:
        This can be used as follows::

            at = AsyncThread()
            future = at.run(my_coro())
            # do some other stuff
            at.shutdown()

    Warning:
        Remember that you **must** always used thread-safe asyncio invocation methods with this
        class.

    """

    _instance = None
    # pylint: disable=protected-access
    def __new__(cls, *args, **kwargs):
        """Implement the Singleton pattern."""
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
            cls._instance._state = 'new'
        else:
            cls._instance._state = 'exists'
        if not hasattr(cls._instance, 'thread'):
            cls._instance.thread = None

        if not hasattr(cls._instance, 'loop'):
            cls._instance.loop = None

        return cls._instance

    def __init__(self) -> None:
        """Create/reinitialise a thread and set up an asyncio event loop in it.

        If the thread/loop already exist, this will make sure that they are running,
        otherwise it will create new ones.

        """
        self.thread: threading.Thread
        self.loop: asyncio.AbstractEventLoop

        def _start_event_loop_thread() -> None:
            """Run an asyncio event loop inside this thread."""
            asyncio.set_event_loop(self.loop)
            LOGGER.info(f'Starting asyncio loop in thread: {self.thread.name}.')
            self.loop.run_forever()

        # pylint: disable=access-member-before-definition
        if self._state == 'new':  # type: ignore
            # set up everything as this is the first invocation.
            self.loop = asyncio.new_event_loop()
            LOGGER.info('Created new event loop.')
            for sig in (signal.SIGINT, signal.SIGTERM):
                self.loop.add_signal_handler(sig,
                                             functools.partial(asyncio.create_task,
                                                               self._shutdown_signal_handler(sig)
                                                               )
                                             )
            LOGGER.debug('Added signal handlers...')
            self.thread = threading.Thread(target=_start_event_loop_thread,
                                           name=__name__,
                                           daemon=True)
            LOGGER.debug(f'Started thread {self.thread.name}.')
            self.thread.start()
            self._state = 'exists'

        else:
            LOGGER.info('Loop already exists, so not creating a new one.')
            if not self.thread or not self.thread.is_alive():
                self.thread = threading.Thread(target=_start_event_loop_thread,
                                               name=__name__,
                                               daemon=True)
                LOGGER.debug(f'Started thread {self.thread.name}.')
                self.thread.start()

        self._shutdown_sentinel = False
        asyncio.run_coroutine_threadsafe(self._loop_monitor(), self.loop)

    def run(self, coro: Coroutine) -> Future:
        """Run a coroutine in the async thread.

        Warning:
            This is a non-blocking method.

        Returns:
            concurrent.futures.Future: a Future holding any results returned from the call.

        """
        return asyncio.run_coroutine_threadsafe(
            coro, self.loop)


    # def __enter__(self):
    #     """Open the context block."""
    #     AsyncThread._context_count += 1
    #     LOGGER.debug(f'Started context block. Nest level = {AsyncThread._context_count}.')
    #     return self

    # def __exit__(self, ex_type, ex_val, traceback):
    #     """Close the context block and shutdown the thread if this is outmost block."""
    #     AsyncThread._context_count -= 1
    #     LOGGER.debug(f'Exited context block. Nest level = {AsyncThread._context_count}.')
    #     if AsyncThread._context_count == 0:
    #         LOGGER.debug(f'Nest level = {AsyncThread._context_count}. Shutting down.')
    #         self.shutdown()


    @property
    def shutdown_sentinel(self):
        """Flag imminent shutdown of the event loop and its thread.

        This sentinel can be used in coroutines that want to loop ``while
        True``. Instead of doing that do::

             while not at.shutdown_sentinel:
                 ...

        Returns:
            shutdown_sentinel (bool): If ``True`` then the event loop has been
            instructed to shut down.

        """
        return self._shutdown_sentinel


    async def _loop_monitor(self) -> None:
        """Monitor the loop."""
        while not self._shutdown_sentinel:
            LOGGER.debug(f'Event loop still running in thread: {self.thread.name}')
            await asyncio.sleep(1)

    async def _cancel_all_tasks(self):
        """Cancel all running tasks in the loop."""
        tasks = [task for task in asyncio.all_tasks(loop=self.loop) if task is not
                 asyncio.current_task(loop=self.loop)]
        for task in tasks:
            task.cancel()
        LOGGER.info(f'Cancelled {len(tasks)} running tasks.')

    # no unit test for the below, but tested manually... Honestly... :)
    async def _shutdown_signal_handler(self, sig: Enum) -> None:  # pragma: no cover
        """Shut down the event loop cleanly."""
        LOGGER.info(f'Caught signal: {sig.name}. Shutting down...')
        self._shutdown_sentinel = True
        await self._cancel_all_tasks()
        self.loop.stop()

    def shutdown(self) -> None:
        """Shutdown the running asyncio thread."""
        LOGGER.info(f'Shutting down...')
        self._shutdown_sentinel = True
        def stop_callback(_):
            self.loop.stop()

        fut = asyncio.run_coroutine_threadsafe(self._cancel_all_tasks(), self.loop)
        fut.add_done_callback(stop_callback)



import asyncio
import logging
import sys
import time

import pytest

import aiocron
from pyskyq import CronThread

logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
logging.basicConfig(level=logging.DEBUG, stream=sys.stdout,
                    format=logformat)  # datefmt="%Y-%m-%d %H:%M:%S"

def test_cronthread(capsys):
    t = CronThread()

    def spam():
        print('It works')


    t.crontab("* * * * * *",
              func=spam,
              start=True
              )

    time.sleep(3)
    t.stop()

    captured = capsys.readouterr()
    assert captured.out == "It works\nIt works\n"



import asyncio
import json
import logging
import sys
import time
from pathlib import Path

import pytest
from asynctest import CoroutineMock, MagicMock

from pyskyq import EPG, Channel, XMLTVListing, channel_from_skyq_service

from .asynccontextmanagermock import AsyncContextManagerMock
from .isloated_filesystem import isolated_filesystem
from .mock_constants import (SERVICE_DETAIL_1, SERVICE_DETAIL_2,
                             SERVICE_SUMMARY_MOCK, FUZZY_CHANNELS_MOCK)

logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
logging.basicConfig(level=logging.WARNING, stream=sys.stdout,
                    format=logformat)  # datefmt="%Y-%m-%d %H:%M:%S"

json_data = ""

def test_EPG_sky_channels(mocker):

    jsonmock_invocation_count = 0

    class jsonmock:

        @staticmethod
        async def json():
            nonlocal jsonmock_invocation_count
            jsonmock_invocation_count += 1
            if jsonmock_invocation_count == 1:
                return json.loads(SERVICE_SUMMARY_MOCK)
            if jsonmock_invocation_count == 2:
                return json.loads(SERVICE_DETAIL_1)
            if jsonmock_invocation_count == 3:
                return json.loads(SERVICE_DETAIL_2)


    client_response = asyncio.Future()
    client_response.set_result(jsonmock)

    a = mocker.patch('aiohttp.ClientSession.get', new_callable=AsyncContextManagerMock)
    a.return_value = client_response

    epg = EPG('test_load_channel_list_fake_host')

    with pytest.raises(ValueError, match='No channels loaded.'):
        epg.get_channel_by_sid(2002)

    epg.load_skyq_channel_data()

    time.sleep(2)
    assert isinstance(epg, EPG)
    assert len(epg._channels) == 2
    assert epg.get_channel_by_sid(2002).c == "101"
    assert epg.get_channel_by_sid(2002).t == "BBC One Lon"
    assert epg.get_channel_by_sid(2002).name == "BBC One Lon"
    assert epg.get_channel_by_sid('2002').c == "101"
    assert epg.get_channel_by_sid('2002').t == "BBC One Lon"
    assert epg.get_channel_by_sid('2002').name == "BBC One Lon"

    assert epg.get_channel_by_sid(2002).isbroadcasting is True
    assert "BBC ONE for Greater London and the surrounding area." in \
        epg.get_channel_by_sid(2002).upgradeMessage

    assert "Dave is the home of witty banter with quizcoms, cars and comedies." in \
        epg.get_channel_by_sid(2306).upgradeMessage

    with pytest.raises(ValueError, match='Sid:1234567 not found.'):
        epg.get_channel_by_sid(1234567)

def test_apply_EPG_XMLTV_listing(mocker):

    # set up the epg object with sky data.
    jsonmock_invocation_count = 0

    class jsonmock:
        @staticmethod
        async def json():
            nonlocal jsonmock_invocation_count
            jsonmock_invocation_count += 1
            if jsonmock_invocation_count == 1:
                return json.loads(SERVICE_SUMMARY_MOCK)
            if jsonmock_invocation_count == 2:
                return json.loads(SERVICE_DETAIL_1)
            if jsonmock_invocation_count == 3:
                return json.loads(SERVICE_DETAIL_2)

    client_response = asyncio.Future()
    client_response.set_result(jsonmock)

    a = mocker.patch('aiohttp.ClientSession.get', new_callable=AsyncContextManagerMock)
    a.return_value = client_response

    epg = EPG('test_load_channel_list_fake_host')


    l = XMLTVListing('http://host.com/some/feed')
    l._full_path = Path(__file__).resolve().parent.joinpath('parse_xmltv_data.xml')

    with pytest.raises(ValueError, match='No XMLTVListing file found.'):
        epg.apply_XMLTVListing(l)

    l._downloaded = True

    epg.load_skyq_channel_data()
    time.sleep(2)

    global json_data
    json_data = epg.as_json()   # save state here so we can use in the next test.

    # now test the xml loading.
    epg.apply_XMLTVListing(l)

    assert epg.get_channel_by_sid(2002).xmltv_id == 'f3932e75f691561adbe3b609369e487b'
    assert epg.get_channel_by_sid(2002).xmltv_display_name == 'BBC One Lon'
    assert epg.get_channel_by_sid(2002).xmltv_icon_url.human_repr() == 'http://www.xmltv.co.uk/images/channels/f3932e75f691561adbe3b609369e487b.png'


xmlfile_path = Path(__file__).resolve().parent.joinpath('parse_xmltv_data.xml')

def test_cronjob(mocker):
    # set up the epg object with sky data.
    epg = EPG('testing_cron')

    global json_data
    epg.from_json(json_data)

    assert epg.get_channel_by_sid(2002).name == "BBC One Lon"

    # now mock the xml server
    class xmlmock:
        status = 200
        headers = {'Last-Modified': 'Mon, 08 Oct 2018 01:50:19 GMT'}


        @staticmethod
        async def read(*args, **kwargs):
            with open(xmlfile_path, 'rb') as fil:
                return fil.read()


    xmlresponse = asyncio.Future()
    xmlresponse.set_result(xmlmock)

    b = mocker.patch('aiohttp.ClientSession.get', new_callable=AsyncContextManagerMock)
    b.return_value = xmlresponse

    # and finally test the cronjob.
    with isolated_filesystem():
        l = XMLTVListing('http://host.com/some/feed')

        l2 = XMLTVListing('http://host.com/some/feed/2')


        with pytest.raises(ValueError, match='Bad cronspec passed.'):
            epg.add_XMLTV_listing_cronjob(l, "Arthur 'Two Sheds' Jackson", run_now=True)


        epg.add_XMLTV_listing_cronjob(l, "1 1 1 1 1", run_now=True)
        epg.add_XMLTV_listing_cronjob(l2, "1 1 1 1 1", run_now=False)


        time.sleep(3)
        assert epg.get_channel_by_sid(2002).xmltv_id == 'f3932e75f691561adbe3b609369e487b'
        assert epg.get_channel_by_sid(2002).xmltv_display_name == 'BBC One Lon'
        assert epg.get_channel_by_sid(2002).xmltv_icon_url.human_repr() == 'http://www.xmltv.co.uk/images/channels/f3932e75f691561adbe3b609369e487b.png'

        with pytest.raises(ValueError, match='XMLTVListing already added.'):
            epg.add_XMLTV_listing_cronjob(l, "1 1 1 1 1", run_now=True)

        time.sleep(2)

        cronjobs = epg.get_cronjobs()
        l, s = cronjobs[0]
        assert len(cronjobs) == 2
        assert isinstance(l, XMLTVListing)
        assert l.url.human_repr() == 'http://host.com/some/feed'
        assert s == "1 1 1 1 1"

        epg.delete_XMLTV_listing_cronjob(l)
        epg.delete_XMLTV_listing_cronjob(l2)



        epg2 = EPG('testing_cron')  # blank epg


        with pytest.raises(ValueError, match='No channels loaded.'):
            epg2.add_XMLTV_listing_cronjob(l, "1 1 1 1 1", run_now=True)

        with pytest.raises(ValueError, match='No cronjob found for the passed XMLTVListing.'):
            epg2.delete_XMLTV_listing_cronjob(l)




import asyncio
import logging
import sys
from datetime import datetime, timezone
from pathlib import Path

import aiohttp
import pytest
from aiohttp import MultipartWriter
from dateutil import tz
from yarl import URL

from pyskyq import XMLTVListing

from .isloated_filesystem import isolated_filesystem

logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
logging.basicConfig(level=logging.DEBUG, stream=sys.stdout,
                    format=logformat)  # datefmt="%Y-%m-%d %H:%M:%S"

LOGGER = logging.getLogger(__name__)

def test_xmltvlisting_init():
    # used to set up and tear down a temp dir for these tests.
    with isolated_filesystem():
        with pytest.raises(TypeError, match='path must be a string or Path object.'):
            l = XMLTVListing('http://blah.com/feed/6715', 6)

        Path.cwd().joinpath('somedir').mkdir()  # test the case where the directory is already there.
        m = XMLTVListing('http://blah.com/feed/6715', 'somedir')
        assert m._url == URL('http://blah.com/feed/6715')
        assert isinstance(m._path, Path)
        assert str(m._path) == 'somedir'
        assert "<XMLTVListing: url='http://blah.com/feed/6715', path='somedir'" in m.__repr__()
        assert m._path.is_dir()
        assert m.url == URL('http://blah.com/feed/6715')

        n = XMLTVListing(URL('http://blah.com/feed/6715'), '.str_path')
        assert n._url == URL('http://blah.com/feed/6715')
        assert isinstance(n._path, Path)
        assert str(n._path) == '.str_path'

        assert m == n
        assert m._filename == n._filename
        assert m.__hash__() == n.__hash__()


xmlfile_path = Path(__file__).resolve().parent.joinpath('fetch_payload.txt')

@pytest.mark.asyncio
async def test_xmltvlisting_fetch_200(aresponses):

    async def get_handler_200(request):
        with open(xmlfile_path, 'r') as fd:
            data = fd.read()
            hdr = {'Last-Modified': 'Mon, 08 Oct 2018 01:50:19 GMT'}
            resp = aresponses.Response(status=200, reason='OK', body=data, headers=hdr)
        return resp


    aresponses.add('foo.com', '/feed/6715', 'get', response=get_handler_200)

    with isolated_filesystem():
        l = XMLTVListing('http://foo.com/feed/6715')
        assert not l._downloaded
        await l.fetch()
        assert l._downloaded
        assert l.file_path.is_file()
        assert l.last_modified == datetime(2018,10,8,1,50,19,0)

        with open(xmlfile_path, 'rb') as src, open(l.file_path, 'rb') as dest:
            assert src.read(-1) == dest.read(-1)


@pytest.mark.asyncio
async def test_xmltvlisting_fetch_206(aresponses):

    async def get_handler_206(request):
        LOGGER.debug(f'request headers = {request.headers}')
        rng = request.http_range
        LOGGER.debug(f'Range = {rng}. Start = {rng.start}. Stop = {rng.stop}. Diff = {rng.stop - rng.start}.')
        with open(xmlfile_path, 'rb') as f:
            f.seek(rng.start)
            data = f.read(rng.stop - rng.start)
            LOGGER.debug(f'data = {data}')
            hdr = {
                'Content-Range': f'bytes {rng.start}-{rng.stop - 1 }/{xmlfile_path.stat().st_size}',
            }
            if rng.stop - 1 > xmlfile_path.stat().st_size:
                LOGGER.debug('Range request went too far...')
                resp = aresponses.Response(status=416, reason='Range Not Satisfiable', body=data)
            else:
                LOGGER.debug('Range request OK.')
                resp = aresponses.Response(status=206, reason='OK', body=data, headers=hdr)
            LOGGER.debug(f'resp header = {hdr}')
        return resp

    for _ in range(1000):
        aresponses.add('foo.com', '/feed/6715', 'get', response=get_handler_206)


    with isolated_filesystem():
        l = XMLTVListing('http://foo.com/feed/6715')
        await l.fetch(range_size=100)
        assert l.file_path.is_file()

        with open(xmlfile_path, 'rb') as src, open(l.file_path, 'rb') as dest:
            assert src.read(-1) == dest.read(-1)

#TODO -- add bad xml file test here.
def test_channel_parse():
    l = XMLTVListing('http://foo.com/feed/6715')

    with pytest.raises(OSError, match='File not downloaded, or download is currently in flight.'):
        for chan in l.parse_channels():
            pass # should throw an error as file is not downloaded.

    l._full_path = Path(__file__).resolve().parent.joinpath('parse_xmltv_data.xml')
    l._downloaded = True

    for i,chan in enumerate(l.parse_channels()):

        LOGGER.debug(chan.xmltv_id)
        LOGGER.debug(chan.xmltv_icon_url)
        LOGGER.debug(chan.xmltv_display_name)

        if i == 0:
            assert chan.xmltv_id == 'f3932e75f691561adbe3b609369e487b'
            assert chan.xmltv_display_name == 'BBC One Lon'
            assert chan.xmltv_icon_url.human_repr() == 'http://www.xmltv.co.uk/images/channels/f3932e75f691561adbe3b609369e487b.png'
        elif i == 1:

            assert chan.xmltv_id == 'a3c70f4c25110a9ca84f7c604023ee6c'
            assert chan.xmltv_display_name == 'Dave'
            assert chan.xmltv_icon_url.human_repr() == 'http://www.xmltv.co.uk/images/channels/a3c70f4c25110a9ca84f7c604023ee6c.png'
        else:
            break



"""Misc util functions."""

import email.utils as eut
import logging
from datetime import datetime
from typing import Any, Iterator
from xml.etree.ElementTree import iterparse

from yarl import URL


def parse_http_date(httpdatetime: str) -> datetime:
    """Turn an HTTP-formated date-time string into a native :py:class:`~datetime.datetime` object.

    Args:
        httpdatetime (str): A string like ``Mon, 08 Oct 2018 01:42:20 GMT``

    Returns:
        datetime: A datetime object.

    """
    return datetime(*eut.parsedate(httpdatetime)[:6])  # type: ignore


# TODO: use  XMLPullParser instead of iterparse() as it's non-blocking...
# TODO: add error checking for XML errors
def xml_parse_and_remove(filename, path) -> Iterator[Any]:
    """Incrementally load and parse an XML file.

    Stolen from Python Cookbook 3rd edition, section 6.4 with credit to the book's authors.
    """
    LOGGER = logging.getLogger(__name__)

    path_parts = path.split('/')
    doc = iterparse(filename, ('start', 'end')) #TODO: see if we can use aiofiles here.
    # skip the root element
    next(doc)  # pylint: disable=stop-iteration-return
    tag_stack = []
    elem_stack = []
    for event, elem in doc:
        if event == 'start':
            LOGGER.debug(f'Parsing XML element: {elem.tag}')
            tag_stack.append(elem.tag)
            elem_stack.append(elem)
        else: # event == 'end'
            if tag_stack == path_parts:
                yield elem
                elem.clear()
            try:
                tag_stack.pop()
                elem_stack.pop()
            except IndexError:
                pass

#TODO: this is ugly, fix at some point.
def skyq_json_decoder_hook(obj):
    """Decode JSON into appropriate types used in the project."""
    if "xmltv_icon_url" in obj.keys():
        obj['xmltv_icon_url'] = URL(obj['xmltv_icon_url'])
    return obj



"""This module implements the EPG class."""

import functools
import json
import logging
from typing import Any, List, Optional, Tuple

from aiohttp import ClientSession, ClientTimeout  # type: ignore
from croniter.croniter import croniter

from dataclasses import dataclass

from .asyncthread import AsyncThread
from .channel import (Channel, _ChannelJSONEncoder, channel_from_json,
                      channel_from_skyq_service, merge_channels)
from .constants import (REST_PORT, REST_SERVICE_DETAIL_URL_PREFIX,
                        REST_SERVICES_URL)
from .cronthread import CronThread
from .xmltvlisting import XMLTVListing

LOGGER = logging.getLogger(__name__)

at = AsyncThread()

class EPG:
    """The top-level class for all EPG data and functions.

    EPG implements access to any data and operations related to the SkyQ box's
    Electronic Programme Guide through a REST endpoint exposed on the box.
    Specifically, it fetches both summary and detail channel information from
    two different endpoints from the box and aggregates this data into a list of
    :class:`~.channel.Channel` objects.

    This channel data is then *augmented* with data from www.xmltv.co.uk which provides
    an XML file which follows the ``xmltv`` DTD.

    It also uses the ``xmltv`` file to load **listings** data to enable the querying of
    programmes that are scheduled on the channels.

    Args:
        host (str): String with resolvable hostname or IPv4 address to SkyQ box.
        rest_port (int): Defaults to the SkyQ REST port which is 9006.

    Attributes:
        host (str): Hostname or IPv4 address of SkyQ Box.
        rest_port (int): Port number to use to connect to the REST HTTP server.
            Defaults to the standard port used by SkyQ boxes which is 9006.

    """

    def __init__(self,
                 host: str,
                 *,
                 rest_port: int = REST_PORT,
                 ) -> None:
        """Initialise Sky EPG Object."""
        self.host: str = host
        self.rest_port: int = rest_port
        self._channels: list = []
        self._jobs: list = []

        #self._loop = loop if loop else asyncio.get_event_loop()

        #assert not self._loop.is_closed()

        LOGGER.debug(f"Initialised EPG object using SkyQ box={self.host}")

    def __repr__(self):
        """Print a human-friendly representation of this object."""
        return f"<EPG: host={self.host}, rest_port={self.rest_port}, " + \
            f"len(_channels)={len(self._channels)}, len(_xmltv_urls)={len(self._channels)}>"

    def channels_loaded(self) -> bool:
        """Return whether channels have been loaded.

        Returns:
            bool: ``True`` if channels have been loaded.

        """
        return bool(self._channels)

    async def _load_channels_from_skyq(self) -> None:
        """Load channel data into channel property.

        This method fetches the channel list from ``/as/services`` endpoint and loads
        :attr:`~.epg.EPG._channels`. It then fetches the detail from ``/as/services/details/<sid>``
        and adds that to the channels.

        Returns:
            None

        """
        url = f'http://{self.host}:{self.rest_port}{REST_SERVICES_URL}'
        LOGGER.debug(f'Fetching channel list from {url}')

        timeout = ClientTimeout(total=60)
        async with ClientSession(timeout=timeout) as session:
            LOGGER.debug(f'About to fetch channel list from sky box.')
            chan_payload_json = await session.get(url)
            LOGGER.debug(f'Got channel list from sky box.')
            chan_payload = await chan_payload_json.json()
            LOGGER.debug(f'Parsed channel list JSON from sky box.')

            for single_channel_payload in chan_payload['services']:
                LOGGER.debug(f'Processing {single_channel_payload}')
                channel = channel_from_skyq_service(single_channel_payload)
                sid = channel.sid
                detail_url = f'http://{self.host}:{self.rest_port}' + \
                    f'{REST_SERVICE_DETAIL_URL_PREFIX}{sid}'
                detail_payload_json = await session.get(detail_url)
                detail_payload = await detail_payload_json.json()
                detailed_channel = channel.load_skyq_detail_data(detail_payload)
                self._channels.append(detailed_channel)


    def as_json(self) -> str:
        """Return the channel and programme data as JSON.

        Returns:
            str: A JSON representation of this EPG.

        """
        return json.dumps(self._channels, cls=_ChannelJSONEncoder, indent=4)

    def from_json(self, json_: str) -> None:
        """Load channel and programme data from JSON.

        Args:
            json_(str): A string of JSON text.

        """
        payload = json.loads(json_)
        new_chans: list = []

        for item in payload:
            if item['__type__'] == "__channel__":
                new_chans.append(channel_from_json(json.dumps(item)))
            # TODO: add jobs, programmes, etc
        self._channels = new_chans


    def load_skyq_channel_data(self) -> None:
        """Load all channel data from the SkyQ REST Service.

        This method loads all channel data from the SkyQ box into the
        :class:`~.epg.EPG` object.

        Warning:
            This method is non-blocking. You may need to wait a bit before
            the channel data is all fully loaded.

        Returns:
            None

        """
        at.run(self._load_channels_from_skyq())

    def get_channel_by_sid(self,
                           sid: Any
                           ) -> Channel:
        """Get channel data by SkyQ service id.

        This method returns a :class:`~.channel.Channel` object when
        passed in a channel ``sid``.

        Args:
            sid: The SkyQ service id of the channel
        Returns:
            :class:`~.channel.Channel`: The channel if found.

        Raises:
            ValueError: If the channel is not found or the channel
                list is empty.

        """
        sid = str(sid)
        if not self.channels_loaded():
            raise ValueError("No channels loaded.")

        for chan in self._channels:
            if chan.sid == sid:
                return chan
        raise ValueError(f"Sid:{sid} not found.")



    @dataclass
    class _CronJob:
        listing: XMLTVListing
        crontab: str
        thread: Optional[CronThread] = None

    def delete_XMLTV_listing_cronjob(self,
                                     listing: XMLTVListing,
                                     ) -> None:
        """Delete an XML TV listing cronjob from the EPG.

        Args:
            listing[XMLTVListing]: An XMLTVListing object to unschedule.

        """
        try:
            cronjob = [job for job in self._jobs if job.listing == listing][0]
        except (ValueError, IndexError) as ve:
            raise ValueError('No cronjob found for the passed XMLTVListing.') from ve

        cronjob.thread.stop()
        del self._jobs[self._jobs.index(cronjob)]


    def add_XMLTV_listing_cronjob(self,
                                  listing: XMLTVListing,
                                  cronspec: str,
                                  *,
                                  run_now: bool = False
                                  ) -> None:
        """Add an XML TV listing cronjob to the EPG.

        This method will add a :class:`~.xmltvlisting.XMLTVListing` to the EPG object
        and immediately schedule it as a cronjob.

        The processing of the XMLTV download and load into memory will be immedately triggered
        if ``run_now`` is ``True``.

        Args:
            listing (XMLTVListing): a :class:`~.xmltvlisting.XMLTVListing` object to
                add to the EPG.
            cronspec (str): cronspec string, e.g.: ``0 9,10 * * * mon,fri``.
            run_now (bool): If ``True`` **runs** the XMLTV job immediately in addition to
                scheduling the cronjob.

        Returns:
            None

        Raises:
            ValueError: Raised if this XMLTV listing is already added to the EPG, the
                cronspec passes is invalid, or the channel list is empty.

        """
        if not self._channels:
            raise ValueError(f"No channels loaded.")

        if listing not in [job.listing for job in self._jobs]:
            if cronspec and croniter.is_valid(cronspec):
                cron_t = CronThread()
                cron_t.crontab(cronspec,
                               func=functools.partial(
                                   self._download_and_apply_XMLTVListing,
                                   listing),
                               start=True
                               )
                cronjob = self._CronJob(listing=listing, crontab=cronspec, thread=cron_t)
                LOGGER.info(f'Listing cronjob added: {cronjob}')
                self._jobs.append(cronjob)
            else:
                raise ValueError('Bad cronspec passed.')
        else:
            raise ValueError('XMLTVListing already added.')

        if run_now:
            at.run(self._download_and_apply_XMLTVListing(listing))


    async def _download_and_apply_XMLTVListing(self, listing: XMLTVListing) -> None:
        """Download a listing and merge its data into the EPG channels data structure.

        This method is intended primarily for use in EPG cronjobs. If you wish to manually
        download and apply an XMLTVListing, you might prefer to call
        :meth:`.xmltvlisting.XMLTVListing.fetch` manually first, followed by
        :meth:`apply_XMLTVListing`.

        Args:
            listing (XMLTVListing): a :class:`~.xmltvlisting.XMLTVListing` object to
                download and apply to the EPG.

        Returns:
            None

        """
        LOGGER.debug(f'Listing = {listing}')
        await listing.fetch()  # do the download
        self.apply_XMLTVListing(listing)


    def apply_XMLTVListing(self, listing: XMLTVListing) -> None:
        """Merge listing data into the EPG channels data structure by channel name.

        Args:
            listing(XMLTVListing): a: class: `~.xmltvlisting.XMLTVListing` object to
            merge to the EPG.

        Returns:
            None

        Raises:
            ValueError: Raised if the XMLTVListing file cannot be found.

        """
        if not listing.downloaded:
            raise ValueError("No XMLTVListing file found.")

        for xmltv_channel in listing.parse_channels():
            sky_channel_names = [chan.name.lower() for chan in self._channels]
            if xmltv_channel.xmltv_display_name.lower() in sky_channel_names:
                idx = sky_channel_names.index(xmltv_channel.xmltv_display_name.lower())
                new_chan = merge_channels(self._channels[idx], xmltv_channel)
                LOGGER.debug(f'New channel:{new_chan}.')
                self._channels.append(new_chan)
                LOGGER.debug(f'Replaced {self._channels[idx]} with {new_chan} in EPG.')
                del self._channels[idx]


    def get_cronjobs(self) -> List[Tuple[XMLTVListing, str]]:
        """Return currently loaded XML TV Listings and associated cronjobs.

        This returns a list of tuples containing
        :class:`~.xmltvlisting.XMLTVListing` objects and associated cronspec
        strings.

        Returns:
            List[Tuple[XMLTVListing, str]]: Returns a list of tuples where each
                tuple is as follows::

                    (XMLTVListing, cronspec)

                where:

                - listing is an :class:`~.xmltvlisting.XMLTVListing` object
                - str is a string like ``0 9,10 * * * mon,fri``

        """
        return [(j.listing, j.crontab) for j in self._jobs]



"""This module implements the XMLTVListing listing class."""
#pylint: disable=line-too-long
import hashlib
import logging
import shutil
from collections.abc import Hashable
from datetime import datetime
from http.client import HTTPException
from pathlib import Path
from typing import Iterator, Optional

from aiohttp import ClientSession, ClientTimeout  # type: ignore
from yarl import URL

from .channel import Channel, channel_from_xmltv_list
from .utils import parse_http_date, xml_parse_and_remove

LOGGER = logging.getLogger(__name__)


class XMLTVListing(Hashable):  # pylint: disable=too-many-instance-attributes
    """Hold the data and functions required to obtain an XMLTVListing.

    Currently it only supports the XML TV Listings format which holds
    Channel and Programme scheduling information. See the dtd_ for details.

    .. _dtd: https://github.com/AlekSi/xmltv/blob/master/xmltv.dtd

    Note:

        The :class:`~.channel.Channel` class provides the **primary**
        interface to channel data through the :class:`~.epg.EPG` class.

        This class provides the means to download and parse XML data to do with
        channels, but more importantly, programming schedules. While it can be used
        stand-alone, it is designed to be injected into the
        :class:`~.epg.EPG` object using
        :meth:`.epg.EPG.add_XMLTV_listing_schedule`.

    """

    def __init__(self,
                 url: URL,
                 path: Path = Path('.epg_data'),
                 ) -> None:
        """Instantiate the object with the URL to fetch."""
        self._url: URL
        self._path: Path

        if not isinstance(url, URL):
            self._url = URL(url)
        else:
            self._url = url

        # Validate path
        if not isinstance(path, Path) and not isinstance(path, str):
            raise TypeError('path must be a string or Path object.')

        # coerce it to a Path if it's a string
        if isinstance(path, str):
            self._path = Path(path)
        else:
            self._path = path

        if not self._path.is_dir():
            self._path.mkdir()

        str_binary = str(self._url.human_repr()).encode('utf8')
        hashobj = hashlib.sha256(str_binary)
        self._hash = int.from_bytes(hashobj.digest(), 'big')
        self._filename = f'{hashobj.hexdigest()}.xml'
        self._full_path = self._path.joinpath(self._filename)
        self._last_modified: Optional[datetime] = None
        self._downloaded: bool = False
        self._downloading: bool = False

        LOGGER.debug(f'XMLTVListing initialised: {self}')

    def __hash__(self) -> int:
        """Define hash function for a Hashable object."""
        return self._hash

    def __eq__(self, other) -> bool:
        """Define equality test for a Hashable object."""
        # relying on lazy boolean evaluation here.
        return isinstance(other, XMLTVListing) and hash(other._url) == hash(self._url)  # pylint: disable=protected-access

    def __repr__(self):
        """Print a human-friendly representation of this object."""
        return f"<XMLTVListing: url='{self._url}', path='{self._path}', filename='{self._filename}'>"

    @property
    def last_modified(self):
        """Return the last modified date from the HTTP header of the last download.

        Returns the date and time when the data was last modified. Taken directly
        from the ``HTTP-Header``.

        Returns:
            datetime: A :py:class:`datetime.datetime` object


        """
        return self._last_modified

    @property
    def url(self) -> URL:
        """Return the url of this XMLTVListing.

        Returns:
            :py:class:`yarl.URL`

        """
        return self._url

    @property
    def downloaded(self) -> bool:
        """Return the status of XMLTV file download.

        Returns:
            bool: ``True`` if the file was downloaded successfully

        """
        return self._downloaded

    @property
    def downloading(self) -> bool:
        """Return the status of XMLTV file download.

        Returns:
            bool: ``True`` if the file is currently being downlaoded

        """
        return self._downloading

    @property
    def file_path(self) -> Path:
        """Return the full file path of this listing's XML file.

        Returns:
            :py:class:`pathlib.Path`: A `Path` to the location of the
            XML file (whether it has yet been :meth:`fetch`'ed or not).

        """
        return self._full_path

    # TODO - add aiofiles support HERE!
    # TODO -- add error handling for
    # -- HTTP headers missing
    # -- timeouts
    # -- etc
    # TODO -- add retry support

    async def fetch(self,  # pylint: disable=too-many-locals
                    *,
                    timeout: int = 60, # sec
                    range_size: int = 256*1024,   # bytes
                    ) -> None:
        """Fetch the XMLTVListing file.

        This async method will download the (XML) file specified by the URL passed
        at instantiation.

        As these files can be large, and because www.xmltv.co.uk, in particular,
        supports Range Requests (see rfc7233_) this method will attempt to download
        the file in parts using HTTP Range Requests, if the server supports them.
        This will limit the use of memory during the download process to that specified
        by the ``range_size`` parameter.

        If the server does not support Range Requests the method will will fall back
        to a more memory-intensive, vanilla HTTP download.

        .. _rfc7233 : https://tools.ietf.org/html/rfc7233

        Args:
            timeout (int): timeout in seconds for the HTTP session. Defaults to
                60 seconds.
            range_size (int): the size, in bytes, of each chunk. Defaults to
                256k (256*1024 bytes).

        Returns:
            None

        """
        LOGGER.debug(f'Fetch({self}) called started.')
        self._downloading = True
        to_ = ClientTimeout(total=timeout)
        async with ClientSession(timeout=to_) as session:
            LOGGER.debug(f'Client session created: {session}. About to fetch url={self._url}')
            byte_start = 0
            byte_stop = range_size - 1  # as we count from 0
            newfile = self._full_path.with_suffix('.tmp')
            with open(newfile, 'wb') as file_desc:  # TODO: use aiofile
                while True:
                    req_header = {"Range": f'bytes={byte_start}-{byte_stop}'}
                    resp = await session.get(self._url, headers=req_header)
                    #async with session.get(self._url, headers=req_header) as resp:
                    # -------
                    LOGGER.debug(f'Attempting byte range download. Range = {byte_start}-{byte_stop}')
                    LOGGER.info(f"Server responsed with status: {resp.status}.")
                    LOGGER.debug(f'Server response headers: {resp.headers}')

                    if 'Last-Modified' in resp.headers:
                        self._last_modified = parse_http_date(resp.headers['Last-Modified'])
                        LOGGER.debug(f'Content last modified on: {resp.headers["Last-Modified"]}.')

                    if resp.status == 416:
                        raise HTTPException("Server responsed with {resp.status}: {resp.message}")  # pragma: no cover

                    if resp.status == 206:  # partial content served.
                        # parse Content-Range header -- it looks like this: Content-Range: bytes 0-1023/16380313
                        content_range_from_total = resp.headers['Content-Range'].split()[1] # remove "bytes" from front of header
                        content_entire_payload_size = int(content_range_from_total.split('/')[1])  # grab the bit after  "/"

                    if resp.status == 200:  # full content served as HTTP server doesn't appear to handle range requests.
                        LOGGER.info("Server responsed with status 200 and not 206 so full download assumed.")

                    chunk = await resp.read()
                    file_desc.write(chunk)
                    if resp.status == 206:
                        if byte_stop + range_size + 1 > content_entire_payload_size: # type: ignore
                            byte_start = byte_stop + 1
                            byte_stop = content_entire_payload_size  # type: ignore
                            if byte_start >= byte_stop:
                                break
                        else:
                            byte_start = byte_stop + 1
                            byte_stop = byte_start + range_size
                        continue
                    break
                    # -------
            shutil.move(newfile, self._full_path)
            self._downloading = False
            self._downloaded = True
        LOGGER.debug(f'Fetch finished on {self}')

    def parse_channels(self) -> Iterator[Channel]:
        """Parse the XMLTVListing XML file and create an iterator over the channels in it."""
        if not self.downloaded and not self.downloading:
            raise OSError('File not downloaded, or download is currently in flight.')
        else:
            LOGGER.debug(f'in parse_channels. file = {self.file_path}')
            for xml_chan in xml_parse_and_remove(self.file_path, 'channel'):
                LOGGER.debug('yielding channel...')
                yield channel_from_xmltv_list(xml_chan)



#!/usr/bin/env python
"""
Cli wrapper for the skyq API.

Example:
    You can invoke the cli client like this:

        ``$ python cli_epg.py pause``

"""

import argparse
import sys
from typing import List
from time import sleep
from pyskyq import EPG, XMLTVListing


def parse_args(args: List[str]) -> argparse.Namespace:
    """Parse command line parameters."""
    parser = argparse.ArgumentParser(
        description="Query Channel Description from SkyQ EPG")
    parser.add_argument(
        dest="sid",
        help="channel sid, try 2002 as an example",
        type=str,
        metavar="SID")
    return parser.parse_args(args)


def main(args: List[str]):
    """Run main routine, allowing arguments to be passed."""
    pargs = parse_args(args)
    epg = EPG('skyq')  # replace with hostname / IP of your Sky box
    epg.load_skyq_channel_data()  # load channel listing from Box.
    all_72_hour = XMLTVListing('http: // www.xmltv.co.uk/feed/6715')
    limited_7_day = XMLTVListing('http://www.xmltv.co.uk/feed/6784')

    sleep(5)


    epg.add_XMLTV_listing_cronjob(limited_7_day, '0 2 * * *', run_now=True)  # at At 02:00 every day
    epg.add_XMLTV_listing_cronjob(all_72_hour, '0 3 * * *', run_now=True)  # at At 03:00 every day


    sec = 0
    while True:
        print(f'Seconds = {sec}')
        print('Channel Description from the SkyQ Box:')
        print(epg.get_channel_by_sid(pargs.sid).desc)
        print('Channel XMLTV ID from the XMLTV Feed:')
        print(epg.get_channel_by_sid(pargs.sid).xmltv_id)
        sleep(1)
        sec += 1
    # for job in epg.get_cronjobs():
    #     epg.delete_XMLTV_listing_cronjob(job[0])

if __name__ == "__main__":
    main(sys.argv[1:])



"""This module implements the Status class.

It schedules subscriptions/polling tasks that monitor various events emitted
from the SkyQ box. Each of these will await an event (e.g. on a websocket or
uPNP subscription) and then update object attributes when neccessary.
"""
import asyncio
import json
import logging
import socket
from typing import Dict

import websockets

from .constants import REST_PORT, REST_STATUS_URL
from .asyncthread import AsyncThread
LOGGER = logging.getLogger(__name__)

at = AsyncThread()
class Status:
    """This class provides real-time access to the status of the SkyQ box.

    It uses an internal :py:mod:`asyncio`-based concurrency object used to
    subscribe to various web-socket and/or uPNP based events.

    It also logs all status changes.

    Args:
        host (str): String with resolvable hostname or IPv4 address to SkyQ box.
        port (int, optional): Port number to use to connect to the Remote REST API.
            Defaults to the standard port used by SkyQ boxes which is ``9006``.
        ws_url_path (str, optional): Path string to append to the URL, defaults to
            ``/as/system/status``
        ws_timeout (int, optional): Web socket connection timeout. Defaults is 20 sec.
        ping_timeout (int, optional): Web socket ping timeout. Defaults is 10 sec.

    Attributes:
        host (str): Hostname or IPv4 address of SkyQ Box.
        port (int): Port number to use to connect to the REST HTTP server.
            Defaults to the standard port used by SkyQ boxes which is 9006.
        ws_url (str): ``ws://`` based URL to the SkyQ Box websocket.
        standby (bool): A Boolean indicator stating whether the box is in
            Standby Mode or not.

    """

    def __init__(self,
                 host: str,
                 *,
                 port: int = REST_PORT,
                 ws_url_path: str = REST_STATUS_URL,
                 ws_timeout: int = 20,
                 ping_timeout: int = 10,
                 ) -> None:
        """Initialise the Status object."""
        self.host: str = host
        self.port: int = port
        self.standby: bool = False
        self.ws_url: str = f'ws://{self.host}:{self.port}{ws_url_path}'

        self._shutdown_sentinel: bool = False # used to trigger a clean shutdown.

        self._ws_timeout: int = ws_timeout
        self._ping_timeout: int = ping_timeout
        LOGGER.debug(f"Initialised Status object object with host={host}, port={port}")


    async def _ws_subscribe(self) -> None:
        """Fetch data from web socket asynchronously.

        This  method fetches data from a web socket asynchronously. It is used to
        subscribe to websockets of the SkyQ box.

        """
        LOGGER.debug(f'Setting up web socket listener on {self.ws_url}.')

        while not at.shutdown_sentinel and not self._shutdown_sentinel:
        # outer loop that will restart every time the connection fails (if sentinel says its okay)
            LOGGER.debug('No shutdown sentinel set, so (re)-starting websocket connection.')
            try:
                async with websockets.connect(self.ws_url) as ws:

                    while not at.shutdown_sentinel and not self._shutdown_sentinel:
                        # listener loop
                        LOGGER.debug('Starting websocket listener loop iteration...')
                        try:
                            LOGGER.debug('Waiting for data...')
                            payload = await asyncio.wait_for(ws.recv(), timeout=self._ws_timeout)
                            LOGGER.debug(f'Web-socket data received. size = {len(payload)}')
                        except (asyncio.TimeoutError,
                                websockets.exceptions.ConnectionClosed) as err:
                            LOGGER.debug(f'Websocket timed out or was closed. Error = {err}')
                            try:
                                if at.shutdown_sentinel or self._shutdown_sentinel:
                                    LOGGER.debug(f'Shutdown triggered. Shutting down...')
                                    break # inner listening loop.
                                else:
                                    LOGGER.debug(f'Trying ping message...')
                                    pong = await ws.ping()
                                    await asyncio.wait_for(pong, timeout=self._ping_timeout)
                                    LOGGER.debug(f'Ping OK, keeping connection alive.')
                            except:  # pylint: disable=bare-except
                                LOGGER.debug(f'Ping timeout - retrying...')
                                break # inner listener loop
                        # process payload
                        LOGGER.debug(f'Invoking payload handler on received message...')
                        asyncio.create_task(self._handle(json.loads(payload)))

            except (socket.gaierror, ConnectionRefusedError) as sc_err:
                await asyncio.sleep(.1)
                LOGGER.debug(f'Could not connect to web socket. Error={sc_err}. Retrying...')
                # sleep a bit, log it, then try again
                continue


    def create_event_listener(self) -> None:
        """Create status event listener coroutine and schedule it.

        This method spawns a coroutine that runs using the facility provided by
        :class:`.asyncthread.AsyncThread`. The coroutine manages all
        subscriptions to various endpoints/services on the SkyQ box and takes
        the appropriate update actions when these events fire.

        This means that you will be able to simply refer to properties like
        :attr:`standby` and rely on them to report the _current_ status from the
        SkyQ Box.

        """
        at.run(self._ws_subscribe())


    async def _handle(self,
                      payload: Dict,
                      ) -> None:
        """Update status properties with payload."""
        self.standby = payload['hdmi']['state'] != 'available'
        LOGGER.info(f'standby =  {str(self.standby)}')


    def shudown_event_listener(self) -> None:
        """Terminate the event listener thread and event loop."""
        self._shutdown_sentinel = True
